# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from . import record as _record
from .meta import InlineMetaclassBase
from six import iteritems
from six.moves import zip

import types
import six


class Constructor(object):
  """
  Represents a constructor for a sumtype. Constructors are declared using
  record types (see the #record module) or by creating #Constructor objects
  from scratch (with the same interface as the #recor.create_record()
  function).

  This class is only used temporarily during the declaration of a sumtype
  and is then replaced by a #BoundConstructor.
  """

  def __init__(self, record_or_fields=(), *mixins):
    """
    Creates a new constructor object.

    # Arguments (1)

    record (type): A #record.CleanRecord subclass.

    # Arguments (2)

    fields (list, dict, str): Fields for the new record class.
    mixins (tuple of type): Mixins for the record subclass.
    """

    if isinstance(record_or_fields, type):
      if not issubclass(record_or_fields, _record.CleanRecord):
        raise TypeError('expected record.CleanRecord subclass')
      record = record_or_fields
    else:
      record = _record.create('_Temporary', record_or_fields, *mixins)
    self.record = record

  def add_member(self, name, value):
    setattr(self.record, name, value)

  def bind(self, name, sumtype):
    name = sumtype.__name__ + '.' + name
    attrs = {'__skip_sumtype_meta__': True}
    typ = type(name, (sumtype, self.record), attrs)
    typ.__constructor__ = self
    return typ


class member_of(object):
  """
  A decorator for functions or values that are supposed to be members of
  only a specific sumtype's constructor (or multiple constructors). Instances
  of this class will be automatically unpacked by the :class:`_SumtypeMeta`
  constructor and moved into the :attr:`Constructor.members` dictionary.
  """

  def __init__(self, constructors=None, value=None, name=None):
    if isinstance(constructors, Constructor):
      constructors = [constructors]
    if any(not isinstance(x, Constructor) for x in constructors):
      raise TypeError('expected only Constructor objects')
    self.constructors = constructors
    self.value = value
    self.name = name
    if self.name:
      self._propagate()

  def __call__(self, value):
    if self.name:
      raise RuntimeError('member_of.name already set')
    self.name = value.__name__
    self.value = value
    self._propagate()
    return self

  def _propagate(self):
    assert self.name
    for c in self.constructors:
      c.add_member(self.name, self.value)


class Sumtype(InlineMetaclassBase):
  """
  Base class for sumtypes.
  """

  __addins__ = []
  __constructors__ = {}

  def __metanew__(cls, name, bases, attrs):
    subtype = type.__new__(cls, name, bases, attrs)
    if attrs.get('__skip_sumtype_meta__', False):
      return subtype

    # Get all previous constructors and get all new ones.
    constructors = getattr(subtype, '__constructors__', {}).copy()
    for key, value in iteritems(vars(subtype)):
      if isinstance(value, Constructor):
        constructors[key] = value
      elif isinstance(value, type) and issubclass(value, _record.CleanRecord):
        constructors[key] = Constructor(value)

    # Update constructors from member_of declarations.
    for key, value in list(iteritems(vars(subtype))):
      if isinstance(value, member_of):
        delattr(subtype, key)

    # Bind constructors.
    for key, value in iteritems(constructors):
      setattr(subtype, key, value.bind(key, subtype))
    subtype.__constructors__ = constructors

    # Invoke addons.
    for addin in getattr(subtype, '__addins__', []):
      addin(subtype)

    return subtype


def add_is_methods(sumtype):
  """
  A sumtype add-in that adds an `is_...()` methods for every constructor.
  """

  import re

  def create_is_check(func_name, constructor_name):
    def check(self):
      return type(self) == getattr(self, constructor_name)
    check.__name__ = func_name
    check.__qualname__ = func_name
    return check

  for name in sumtype.__constructors__.keys():
    func_name = 'is_' + '_'.join(re.findall('[A-Z]+[^A-Z]*', name)).lower()
    setattr(sumtype, func_name, create_is_check(func_name, name))


Sumtype.constructor = Constructor
Sumtype.member_of = member_of
Sumtype.add_is_methods = add_is_methods
Sumtype.__addins__.append(add_is_methods)

import sys
Sumtype.__module_object__ = sys.modules[__name__]  # keep explicit reference for Py2
sys.modules[__name__] = Sumtype



# -*- coding: utf-8 -*-
# Copyright 2007 Pallets
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
    werkzeug.local
    ~~~~~~~~~~~~~~

    This module implements context-local objects.

    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import copy
from six import PY2


class LocalProxy(object):
    """Acts as a proxy for a werkzeug local.  Forwards all operations to
    a proxied object.  The only operations not supported for forwarding
    are right handed operands and any kind of assignment.

    Example usage::

        from werkzeug.local import Local
        l = Local()

        # these are proxies
        request = l('request')
        user = l('user')


        from werkzeug.local import LocalStack
        _response_local = LocalStack()

        # this is a proxy
        response = _response_local()

    Whenever something is bound to l.user / l.request the proxy objects
    will forward all operations.  If no object is bound a :exc:`RuntimeError`
    will be raised.

    To create proxies to :class:`Local` or :class:`LocalStack` objects,
    call the object as shown above.  If you want to have a proxy to an
    object looked up by a function, you can (as of Werkzeug 0.6.1) pass
    a function to the :class:`LocalProxy` constructor::

        session = LocalProxy(lambda: get_current_request().session)

    .. versionchanged:: 0.6.1
       The class can be instantiated with a callable as well now.
    """

    __slots__ = ("__local", "__dict__", "__name__", "__wrapped__")

    def __init__(self, local, name=None):
        object.__setattr__(self, "_LocalProxy__local", local)
        object.__setattr__(self, "__name__", name)
        if callable(local) and not hasattr(local, "__release_local__"):
            # "local" is a callable that is not an instance of Local or
            # LocalManager: mark it as a wrapped function.
            object.__setattr__(self, "__wrapped__", local)

    def _get_current_object(self):
        """Return the current object.  This is useful if you want the real
        object behind the proxy at a time for performance reasons or because
        you want to pass the object into a different context.
        """
        if not hasattr(self.__local, "__release_local__"):
            return self.__local()
        try:
            return getattr(self.__local, self.__name__)
        except AttributeError:
            raise RuntimeError("no object bound to %s" % self.__name__)

    @property
    def __dict__(self):
        try:
            return self._get_current_object().__dict__
        except RuntimeError:
            raise AttributeError("__dict__")

    def __repr__(self):
        try:
            obj = self._get_current_object()
        except RuntimeError:
            return "<%s unbound>" % self.__class__.__name__
        return repr(obj)

    def __bool__(self):
        try:
            return bool(self._get_current_object())
        except RuntimeError:
            return False

    if PY2:
        __nonzero__ = __bool__
        del __bool__

    def __unicode__(self):
        try:
            return unicode(self._get_current_object())  # noqa
        except RuntimeError:
            return repr(self)

    def __dir__(self):
        try:
            return dir(self._get_current_object())
        except RuntimeError:
            return []

    def __getattr__(self, name):
        if name == "__members__":
            return dir(self._get_current_object())
        return getattr(self._get_current_object(), name)

    def __setitem__(self, key, value):
        self._get_current_object()[key] = value

    def __delitem__(self, key):
        del self._get_current_object()[key]

    if PY2:
        __getslice__ = lambda x, i, j: x._get_current_object()[i:j]

        def __setslice__(self, i, j, seq):
            self._get_current_object()[i:j] = seq

        def __delslice__(self, i, j):
            del self._get_current_object()[i:j]

    __setattr__ = lambda x, n, v: setattr(x._get_current_object(), n, v)
    __delattr__ = lambda x, n: delattr(x._get_current_object(), n)
    __str__ = lambda x: str(x._get_current_object())
    __lt__ = lambda x, o: x._get_current_object() < o
    __le__ = lambda x, o: x._get_current_object() <= o
    __eq__ = lambda x, o: x._get_current_object() == o
    __ne__ = lambda x, o: x._get_current_object() != o
    __gt__ = lambda x, o: x._get_current_object() > o
    __ge__ = lambda x, o: x._get_current_object() >= o
    __cmp__ = lambda x, o: cmp(x._get_current_object(), o)  # noqa
    __hash__ = lambda x: hash(x._get_current_object())
    __call__ = lambda x, *a, **kw: x._get_current_object()(*a, **kw)
    __len__ = lambda x: len(x._get_current_object())
    __getitem__ = lambda x, i: x._get_current_object()[i]
    __iter__ = lambda x: iter(x._get_current_object())
    __contains__ = lambda x, i: i in x._get_current_object()
    __add__ = lambda x, o: x._get_current_object() + o
    __sub__ = lambda x, o: x._get_current_object() - o
    __mul__ = lambda x, o: x._get_current_object() * o
    __floordiv__ = lambda x, o: x._get_current_object() // o
    __mod__ = lambda x, o: x._get_current_object() % o
    __divmod__ = lambda x, o: x._get_current_object().__divmod__(o)
    __pow__ = lambda x, o: x._get_current_object() ** o
    __lshift__ = lambda x, o: x._get_current_object() << o
    __rshift__ = lambda x, o: x._get_current_object() >> o
    __and__ = lambda x, o: x._get_current_object() & o
    __xor__ = lambda x, o: x._get_current_object() ^ o
    __or__ = lambda x, o: x._get_current_object() | o
    __div__ = lambda x, o: x._get_current_object().__div__(o)
    __truediv__ = lambda x, o: x._get_current_object().__truediv__(o)
    __neg__ = lambda x: -(x._get_current_object())
    __pos__ = lambda x: +(x._get_current_object())
    __abs__ = lambda x: abs(x._get_current_object())
    __invert__ = lambda x: ~(x._get_current_object())
    __complex__ = lambda x: complex(x._get_current_object())
    __int__ = lambda x: int(x._get_current_object())
    __long__ = lambda x: long(x._get_current_object())  # noqa
    __float__ = lambda x: float(x._get_current_object())
    __oct__ = lambda x: oct(x._get_current_object())
    __hex__ = lambda x: hex(x._get_current_object())
    __index__ = lambda x: x._get_current_object().__index__()
    __coerce__ = lambda x, o: x._get_current_object().__coerce__(x, o)
    __enter__ = lambda x: x._get_current_object().__enter__()
    __exit__ = lambda x, *a, **kw: x._get_current_object().__exit__(*a, **kw)
    __radd__ = lambda x, o: o + x._get_current_object()
    __rsub__ = lambda x, o: o - x._get_current_object()
    __rmul__ = lambda x, o: o * x._get_current_object()
    __rdiv__ = lambda x, o: o / x._get_current_object()
    if PY2:
        __rtruediv__ = lambda x, o: x._get_current_object().__rtruediv__(o)
    else:
        __rtruediv__ = __rdiv__
    __rfloordiv__ = lambda x, o: o // x._get_current_object()
    __rmod__ = lambda x, o: o % x._get_current_object()
    __rdivmod__ = lambda x, o: x._get_current_object().__rdivmod__(o)
    __copy__ = lambda x: copy.copy(x._get_current_object())
    __deepcopy__ = lambda x, memo: copy.deepcopy(x._get_current_object(), memo)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

__all__ = [
  'Interface', 'Implementation', 'ImplementationError', 'ConflictingInterfacesError',
  'is_interface', 'implements', 'attr', 'default', 'final', 'override'
]

import itertools
import six
import sys
import types

from . import NotSet
from .meta import InlineMetaclassBase


class _Member(object):

  def __init__(self, interface, name, hidden=False):
    self.interface = interface
    self.name = name
    self.hidden = hidden

  def __repr__(self):
    result = '<{} {!r}'.format(type(self).__name__, self.name)
    if self.interface:
      result += ' of interface {!r}'.format(self.interface.__name__)
    return result + '>'

  @property
  def is_bound(self):
    if self.interface and self.name:
      return True
    return False


class Method(_Member):

  def __init__(self, interface, name, impl=None, final=False, hidden=False):
    super(Method, self).__init__(interface, name)
    self.impl = impl
    self.final = final
    self.hidden = hidden

  def __repr__(self):
    s = super(Method, self).__repr__()
    if self.hidden:
      s = '<hidden ' + s[1:]
    if self.final:
      s = '<final ' + s[1:]
    return s

  def __call__(self, *a, **kw):
    if self.impl:
      return self.impl(*a, **kw)
    return None

  @classmethod
  def is_candidate(cls, name, value):
    if name.startswith('_') and not name.endswith('_'):  # Private function
      return False
    return isinstance(value, types.FunctionType)

  @classmethod
  def wrap_candidate(cls, interface, name, value):
    if cls.is_candidate(name, value):
      # We don't want these functions to be an actual "member" of the interface
      # API as they can be implemented for every interface and not collide.
      hidden = name in ('__new__', '__init__', '__constructed__')
      # If it's one of the hidden methods, they also act as the "default"
      # implementation because we actually want to call them independently
      # of potential overrides by the implementation.
      impl = value if getattr(value, '__is_default__', hidden) else None
      final = getattr(value, '__is_final__', False)
      return Method(interface, name, impl, final, hidden)
    return None


class Attribute(_Member):
  """
  Represents an attribute on an interface. Note that attributes on interface
  can conflict the same way methods can do. Usually, attribute declaratons
  are only used if the interface adds the respective member in `__init__()`.

  Inside an interface declaration, use the #attr() function to create an
  attribute that will be bound automatically when the interface class is
  constructed.
  """

  def __init__(self, interface, name, type=None):
    super(Attribute, self).__init__(interface, name)
    self.type = type


class Property(_Member):
  """
  Represents a property in an interface. A property can have default
  implementations for the getter, setter and deleter independently.
  """

  def __init__(self, interface, name, getter_impl=None, setter_impl=NotImplemented,
               deleter_impl=NotImplemented, getter_final=False, setter_final=False,
               deleter_final=False):
    super(Property, self).__init__(interface, name)
    self.getter_impl = getter_impl
    self.setter_impl = setter_impl
    self.deleter_impl = deleter_impl
    self.getter_final = getter_final
    self.setter_final = setter_final
    self.deleter_final = deleter_final

  def is_pure_default(self):
    return all(x is not None for x in [self.getter_impl, self.setter_impl, self.deleter_impl])

  def satisfy(self, value):
    assert isinstance(value, property), type(value)
    if value.fget and self.getter_final:
      raise ValueError('propery {}: getter must not be implemented'.format(self.name))
    if value.fset and self.setter_final:
      raise ValueError('propery {}: setter must not be implemented'.format(self.name))
    if value.fdel and self.deleter_final:
      raise ValueError('propery {}: deleter must not be implemented'.format(self.name))
    if self.getter_impl is None and not value.fget:
      raise ValueError('property {}: missing getter'.format(self.name))
    if self.setter_impl is None and not value.fset:
      raise ValueError('property {}: missing setter'.format(self.name))
    if self.deleter_impl is None and not value.fdel:
      raise ValueError('property {}: missing deleter'.format(self.name))

    getter, setter, deleter = value.fget, value.fset, value.fdel
    if not getter and self.getter_impl not in (None, NotImplemented):
      getter = self.getter_impl
    if not setter and self.setter_impl not in (None, NotImplemented):
      setter = self.setter_impl
    if not deleter and self.deleter_impl not in (None, NotImplemented):
      deleter = self.deleter_impl

    return property(getter, setter, deleter)

  @property
  def getter(self):
    return property().getter

  @property
  def setter(self):
    return property().setter

  @property
  def deleter(self):
    return property().deleter

  @classmethod
  def is_candidate(cls, name, value):
    return isinstance(value, property)

  @classmethod
  def wrap_candidate(cls, interface, name, value):
    if cls.is_candidate(name, value):
      return Property.from_python_property(interface, name, value)
    return None

  @classmethod
  def from_python_property(cls, interface, name, value):
    assert isinstance(value, property), type(value)
    if value.fget and getattr(value.fget, '__is_default__', False):
      getter = value.fget
    else:
      getter = None
    if value.fset and getattr(value.fset, '__is_default__', False):
      setter = value.fset
    elif value.fset:
      setter = None
    else:
      setter = NotImplemented
    if value.fdel and getattr(value.fdel, '__is_default__', False):
      deleter = value.fdel
    elif value.fdel:
      deleter = None
    else:
      deleter = NotImplemented
    getter_final = getattr(value.fget, '__is_final__', False)
    setter_final = getattr(value.fset, '__is_final__', False)
    deleter_final = getattr(value.fdel, '__is_final__', False)
    return cls(interface, name, getter, setter, deleter, getter_final,
      setter_final, deleter_final)


class InterfaceClass(type):
  """
  The class for interfaces. Interfaces behave similar to dictionaries.
  """

  def __new__(cls, name, bases, attrs):
    self = type.__new__(cls, name, bases, attrs)
    self.__implementations = set()
    self.__members = {}

    # Convert function declarations in the class to Method objects and
    # bind Attribute objects to the new interface class.
    for key, value in vars(self).items():
      member = None
      if isinstance(value, _Member) and not value.is_bound:
        value.interface = self
        value.name = key
        member = value
      if member is None:
        member = Method.wrap_candidate(self, key, value)
      if member is None:
        member = Property.wrap_candidate(self, key, value)
      if member is not None:
        self.__members[key] = member

    for key in self.__members:
      delattr(self, key)

    return self

  def __contains__(self, key):
    return key in self.__members

  def __getitem__(self, key):
    return self.__members[key]

  def __iter__(self):
    return iter(self.__members)

  def get(self, key, default=None):
    return self.__members.get(key, default)

  def members(self, include_hidden=False):
    for member in six.itervalues(self.__members):
      if include_hidden or not member.hidden:
        yield member

  def implemented_by(self, x):
    if not issubclass(x, Implementation):
      return False
    for interface in x.__implements__:
      if issubclass(interface, self):
        return True
    return False

  def provided_by(self, x):
    if not isinstance(x, Implementation):
      return False
    return self.implemented_by(type(x))

  def implementations(self):
    return iter(self.__implementations)


class Interface(six.with_metaclass(InterfaceClass)):
  """
  Base class for interfaces. Interfaces can not be instantiated.
  """

  def __new__(cls):
    msg = 'interface {} can not be instantiated'.format(cls.__name__)
    raise RuntimeError(msg)


def is_interface(obj):
  return isinstance(obj, type) and issubclass(obj, Interface)


def get_conflicting_members(a, b):
  """
  Returns a set of members that are conflicting between the two interfaces
  *a* and *b*. If the interfaces have no incompatible members, an empty set
  is returned and both interfaces can be implemented in the same
  implementation.
  """

  if not is_interface(a) or not is_interface(b):
    raise TypeError('expected Interface subclass')
  if issubclass(a, b) or issubclass(b, a):
    return set()

  conflicts = []
  for am in a.members():
    try:
      bm = b[am.name]
    except KeyError:
      continue
    if am is not bm:
      conflicts.append(am.name)

  return conflicts


def check_conflicting_interfaces(interfaces):
  """
  Raises a #ConflictingInterfacesError if any of the specified interfaces
  have conflicting members.
  """

  for x in interfaces:
    for y in interfaces:
      if x is not y and get_conflicting_members(x, y):
        raise ConflictingInterfacesError(x, y)


def reduce_interfaces(interfaces):
  """
  Reduces a list of interfaces eliminating classes that are parents of
  other classes in the list.
  """

  result = []
  for interface in interfaces:
    skip = False

    for i in range(len(result)):
      if issubclass(interface, result[i]):
        result[i] = interface
        skip = True
        break
      if issubclass(result[i], interface):
        skip = True
        break

    if not skip:
      result.append(interface)

  return result


class Implementation(InlineMetaclassBase):
  """
  Parent for classes that implement one or more interfaces.
  """

  def __metanew__(cls, name, bases, attrs):
    implements = attrs.setdefault('__implements__', [])
    implements = reduce_interfaces(implements)
    check_conflicting_interfaces(implements)

    # Assign default implementations.
    for interface in implements:
      for member in interface.members():
        if isinstance(member, Method) and member.name not in attrs and member.impl:
          attrs[member.name] = member.impl

    self = type.__new__(cls, name, bases, attrs)

    # Ensure all interface members are satisfied.
    for interface in implements:
      errors = []
      for member in interface.members():
        value = getattr(self, member.name, NotSet)
        if isinstance(member, Method):
          if isinstance(value, types.MethodType):
            value = value.im_func
          if member.final and value is not NotSet and member.impl != value:
            errors.append('implemented final method: {}()'.format(member.name))
            continue
          if value is NotSet:
            errors.append('missing method: {}()'.format(member.name))
          elif not isinstance(value, (types.FunctionType, types.MethodType)):
            errors.append('expected method, got {}: {}()'.format(
              type(value).__name__, member.name))
        elif isinstance(member, Property):
          if not hasattr(member.name):
            if not member.is_pure_default():
              errors.append('missing property: {}'.format(member.name))
          elif not isinstance(value, property):
            errors.append('expected property, got {}: {}'.format(
              type(value).__name__, member.name))
          else:
            try:
              value = member.satisfy(value)
            except ValueError as exc:
              errors.append(str(exc))
            else:
              setattr(self, member.name, value)
      if errors:
        raise ImplementationError(self, interface, errors)

    # Check member functions for whether they have been marked with
    # the @override decorator.
    for key, value in vars(self).items():
      if not isinstance(value, types.FunctionType):
        continue
      if not getattr(value, '__is_override__', False):
        continue
      for interface in implements:
        if key in interface:
          break
      else:
        raise RuntimeError("'{}' does not override a method of any of the "
          "implemented interfaces.".format(key))

    # The implementation is created successfully, add it to the
    # implementations set of all interfaces and their parents.
    for interface in implements:
      bases = [interface]
      while bases:
        new_bases = []
        for x in bases:
          if issubclass(x, Interface):
            x._InterfaceClass__implementations.add(self)
          new_bases += x.__bases__
        bases = new_bases

    return self

  def __init__(self):
    for interface in self.__implements__:
      member = interface.get('__init__')
      if member:
        member.impl(self)
    for interface in self.__implements__:
      member = interface.get('__constructed__')
      if member:
        member.impl(self)


class ImplementationError(RuntimeError):

  def __init__(self, impl, interface, errors):
    self.impl = impl
    self.interface = interface
    self.errors = errors

  def __str__(self):
    lines = []
    lines.append("'{}' does not meet requirements "
                "of interface '{}'".format(self.impl.__name__, self.interface.__name__))
    lines += ['  - {}'.format(x) for x in self.errors]
    return '\n'.join(lines)


class ConflictingInterfacesError(RuntimeError):

  def __init__(self, a, b):
    self.a = a
    self.b = b

  def __str__(self):
    lines = ["'{}' conflicts with '{}'".format(self.a.__name__, self.b.__name__)]
    for member in get_conflicting_members(self.a, self.b):
      lines.append('  - {}'.format(member))
    return '\n'.join(lines)


def implements(*interfaces):
  """
  Decorator for a class to mark it as implementing the specified *interfaces*.
  Note that this will effectively create a copy of the wrapped class that
  inherits from the #Implementation class.
  """

  def decorator(cls):
    attrs = vars(cls).copy()
    attrs.pop('__weakref__', None)
    attrs.pop('__dict__', None)
    attrs['__implements__'] = interfaces
    if cls.__bases__ == (object,):
      bases = (Implementation,)
    else:
      bases = cls.__bases__ + (Implementation,)
    return type(cls.__name__, bases, attrs)

  return decorator


def attr(type=None):
  """
  Declare an unnamed attribute that will be bound when the interface is
  constructed. The result of this function must be assigned to a member
  on the class-level of an #Interface declaration.
  """

  return Attribute(None, None, type)


def default(func):
  """
  Decorator for interface methods to mark them as a default implementation.
  """

  func.__is_default__ = True
  return func


def final(func):
  """
  Decorator for an interface method or property component to mark it as a
  default implementation and that it may not actually be implemented.
  """

  func.__is_default__ = True
  func.__is_final__ = True
  return func


def override(func):
  """
  Marks a function as expected override a method in an implemented interface.
  If the function does not override a method in an implemented interface,
  a #RuntimeError will be raised when the #Implementation subclass is created.
  """

  func.__is_override__ = True
  return func



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import six

from nr.types.abc import MutableMapping


class OrderedDict(MutableMapping):
  """
  This class implements a dictionary that can treat non-hashable
  datatypes as keys. It is implemented as a list of key/value pairs,
  thus it is very slow compared to a traditional hash-based dictionary.

  Access times are:

  ==== ======= =====
  Best Average Worst
  ==== ======= =====
  O(1) O(n)    O(n)
  ==== ======= =====
  """

  def __init__(self, iterable=None):
    super(OrderedDict, self).__init__()
    if isinstance(iterable, OrderedDict):
      self.__items = iterable.__items[:]
    elif iterable is not None:
      self.__items = []
      for key, value in iterable:
        self.__items.append((key, value))
    else:
      self.__items = []

  __hash__ = None

  def __contains__(self, key):
    for item in self.__items:
      if item[0] == key:
        return True
    return False

  def __eq__(self, other):
    if other is self:
      return True
    return other == self.asdict()

  def __ne__(self, other):
    return not self == other

  def __len__(self):
    return len(self.__items)

  def __str__(self):
    items = ('{0!r}: {1!r}'.format(k, v) for k, v in self.__items)
    return '{' + ', '.join(items) + '}'

  __repr__ = __str__

  def __iter__(self):
    for item in self.__items:
      yield item[0]

  def __getitem__(self, key):
    for item in self.__items:
      if item[0] == key:
        return item[1]
    raise KeyError(key)

  def __setitem__(self, key, value):
    for item in self.__items:
      if item[0] == key:
        item[1] = value
        return
    self.__items.append([key, value])

  def __delitem__(self, key):
    for index, item in enumerate(self.__items):
      if item[0] == key:
        break
    else:
      raise KeyError(key)

    del self.__items[index]

  def iterkeys(self):
    for item in self.__items:
      yield item[0]

  def itervalues(self):
    for item in self.__items:
      yield item[1]

  def iteritems(self):
    for item in self.__items:
      yield (item[0], item[1])

  if six.PY2:
    def keys(self):
      return list(self.iterkeys())

    def values(self):
      return list(self.itervalues())

    def items(self):
      return list(self.iteritems())
  else:
    keys = iterkeys
    values = itervalues
    items = iteritems
    del iterkeys, itervalues, iteritems

  def get(self, key, default=None):
    try:
      return self[key]
    except KeyError:
      return default

  def pop(self, key, default=NotImplementedError):
    for item in self.__items:
      if item[0] == key:
        break
    else:
      if default is NotImplementedError:
        raise KeyError(key)
      return default

    return key[1]

  def popitem(self, last=True):
    if last:
      return tuple(self.__items.pop())
    else:
      return tuple(self.__items.pop(0))

  def setdefault(self, key, value=None):
    for item in self.__items:
      if item[0] == key:
        return item[1]
    self.__items.append([key, value])
    return value

  def update(self, __data__=None, **kwargs):
    if __data__ is not None:
      data = dict(data)
      for key, value in data.iteritems():
        self[key] = value
    for key, value in kwargs.iteritems():
      self[key] = value

  def clear(self):
    self.__items[:] = []

  def copy(self):
    return OrderedDict(self)

  has_key = __contains__

  def update(self, mapping):
    for key, value in mapping.iteritems():
      self[key] = value

  def sort(self, cmp=None, key=None, reverse=False):
    if key is None:
      key = lambda x: x[0]
    self.__items.sort(key=key)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

__all__ = ['OrderedDict', 'ObjectAsDict', 'ObjectFromDict', 'ChainDict',
           'HashDict', 'ValueIterableDict']

import six

from . import abc, generic

try:
  from collections import OrderedDict
except ImportError:
  from ._impl.ordereddict import OrderedDict

if six.PY2:
  _can_iteritems = lambda x: hasattr(x, 'iteritems')
  _can_iterkeys = lambda x: hasattr(x, 'keys')
else:
  _can_iteritems = lambda x: hasattr(x, 'items')
  _can_iterkeys = lambda x: hasattr(x, 'keys')


class ObjectAsDict(abc.MutableMapping):
  """
  This class wraps an object and exposes its members as mapping.
  """

  def __new__(cls, obj):
    if isinstance(obj, ObjectFromDict):
      return obj._ObjectFromDict__mapping
    return super(ObjectAsDict, cls).__new__(cls)

  def __init__(self, obj):
    self.__obj = obj

  def __repr__(self):
    return 'ObjectAsDict({!r})'.format(self.__obj)

  def __iter__(self):
    return self.keys()

  def __len__(self):
    return len(dir(self.__obj))

  def __contains__(self, key):
    return hasattr(self.__obj, key)

  def __getitem__(self, key):
    try:
      return getattr(self.__obj, key)
    except AttributeError:
      raise KeyError(key)

  def __setitem__(self, key, value):
    setattr(self.__obj, key, value)

  def __delitem__(self, key):
    delattr(self.__obj, key)

  def keys(self):
    return iter(dir(self.__obj))

  def values(self):
    return (getattr(self.__obj, k) for k in dir(self.__obj))

  def items(self):
    return ((k, getattr(self.__obj, k)) for k in dir(self.__obj))

  def get(self, key, default=None):
    return getattr(self.__obj, key, default)

  def setdefault(self, key, value):
    try:
      return getattr(self.__obj, key)
    except AttributeError:
      setattr(self.__obj, key, value)
      return value


class ObjectFromDict(object):
  """
  This class wraps a dictionary and exposes its values as members.
  """

  def __new__(cls, mapping, name=None):
    if isinstance(mapping, ObjectAsDict):
      return mapping._ObjectAsDict__obj
    return super(ObjectFromDict, cls).__new__(cls)

  def __init__(self, mapping, name=None):
    self.__mapping = mapping
    self.__name = name

  def __getattribute__(self, key):
    if key.startswith('_ObjectFromDict__'):
      return super(ObjectFromDict, self).__getattribute__(key)
    try:
      return self.__mapping[key]
    except KeyError:
      raise AttributeError(key)

  def __setattr__(self, key, value):
    if key.startswith('_ObjectFromDict__'):
      super(ObjectFromDict, self).__setattr__(key, value)
    else:
      self.__mapping[key] = value

  def __delattr__(self, key):
    if key.startswith('_ObjectFromDict__'):
      super(ObjectFromDict, self).__delattr__(key)
    else:
      del self.__mapping[key]

  def __dir__(self):
    return sorted(self.__mapping.keys())

  def __repr__(self):
    if self.__name:
      return '<ObjectFromDict name={!r}>'.format(self.__name)
    else:
      return '<ObjectFromDict {!r}>'.format(self.__mapping)


class ChainDict(abc.MutableMapping):
  """
  A dictionary that wraps a list of dictionaries. The dictionaries passed
  into the #ChainDict will not be mutated. Setting and deleting values will
  happen on the first dictionary passed.
  """

  def __init__(self, *dicts):
    if not dicts:
      raise ValueError('need at least one argument')
    self._major = dicts[0]
    self._dicts = list(dicts)
    self._deleted = set()
    self._in_repr = False

  def __contains__(self, key):
    if key not in self._deleted:
      for d in self._dicts:
        if key in d:
          return True
    return False

  def __getitem__(self, key):
    if key not in self._deleted:
      for d in self._dicts:
        try: return d[key]
        except KeyError: pass
    raise KeyError(key)

  def __setitem__(self, key, value):
    self._major[key] = value
    self._deleted.discard(key)

  def __delitem__(self, key):
    if key not in self:
      raise KeyError(key)
    self._major.pop(key, None)
    self._deleted.add(key)

  def __iter__(self):
    return six.iterkeys(self)

  def __len__(self):
    return sum(1 for x in self.keys())

  def __repr__(self):
    if self._in_repr:
      return 'ChainDict(...)'
    else:
      self._in_repr = True
      try:
        return 'ChainDict({})'.format(dict(six.iteritems(self)))
      finally:
        self._in_repr = False

  def __eq__(self, other):
    return dict(self.items()) == other

  def __ne__(self, other):
    return not (self == other)

  def get(self, key, default=None):
    try:
      return self[key]
    except KeyError:
      return default

  def pop(self, key, default=NotImplemented):
    if key not in self:
      if default is NotImplemented:
        raise KeyError(key)
      return default
    self._major.pop(key, None)
    self._deleted.add(key)

  def popitem(self):
    if self._major:
      key, value = self._major.popitem()
      self._deleted.add(key)
      return key, value
    for d in self._dicts:
      for key in six.iterkeys(d):
        if key not in self._deleted:
          self._deleted.add(key)
          return key, d[key]
    raise KeyError('popitem(): dictionary is empty')

  def clear(self):
    self._major.clear()
    self._deleted.update(six.iterkeys(self))

  def copy(self):
    return type(self)(*self._dicts[1:])

  def setdefault(self, key, value=None):
    try:
      return self[key]
    except KeyError:
      self[key] = value
      return value

  def update(self, E, *F):
    if _can_iteritems(E):
      for k, v in six.iteritems(E):
        self[k] = v
    elif _can_iterkeys(E):
      for k in six.iterkeys(E):
        self[k] = E[k]
    elif hasattr(E, 'items'):
      E = E.items()
    for k, v in E:
      self[k] = v
    for Fv in F:
      for k, v in six.iteritems(Fv):
        self[k] = v

  def keys(self):
    seen = set()
    for d in self._dicts:
      for key in six.iterkeys(d):
        if key not in seen and key not in self._deleted:
          yield key
          seen.add(key)

  def values(self):
    seen = set()
    for d in self._dicts:
      for key, value in six.iteritems(d):
        if key not in seen and key not in self._deleted:
          yield value
          seen.add(key)

  def items(self):
    seen = set()
    for d in self._dicts:
      for key, value in six.iteritems(d):
        if key not in seen and key not in self._deleted:
          yield key, value
          seen.add(key)

  if six.PY2:
    iterkeys = keys
    keys = lambda self: list(self.iterkeys())
    itervalues = values
    values = lambda self: list(self.itervalues())
    iteritems = items
    items = lambda self: list(self.iteritems())


class HashDict(generic.Generic['key_hash']):
  """
  This dictionary type can be specialized by specifying a hash function,
  allowing it to be used even with unhashable types as keys.

  Example 1: Specialization through subclassing

  ```python
  class MyDict(HashDict):
    def key_hash(self, key):
      return ...
  ```

  Example 2: Explicit type specialization

  ```
  def my_key_hash(x):
    return ...
  MyDict = HashDict[my_key_hash]
  ```
  ```
  """

  class KeyWrapper(object):
    def __init__(self, key, hash_func):
      self.key = key
      self.hash_func = hash_func
    def __repr__(self):
      return repr(self.key)
    def __hash__(self):
      return self.hash_func(self.key)
    def __eq__(self, other):
      return self.key == other.key
    def __ne__(self, other):
      return self.key != other.key

  def __init__(self):
    generic.assert_initialized(self)
    self._dict = {}

  def __repr__(self):
    return repr(self._dict)

  def __getitem__(self, key):
    key = self.KeyWrapper(key, self.key_hash)
    return self._dict[key]

  def __setitem__(self, key, value):
    key = self.KeyWrapper(key, self.key_hash)
    self._dict[key] = value

  def __delitem__(self, key):
    key = self.KeyWrapper(key, self.key_hash)
    del self._dict[key]

  def __iter__(self):
    return self.iterkeys()

  def __contains__(self, key):
    return self.KeyWrapper(key, self.key_hash) in self._dict

  def items(self):
    return list(self.iteritems())

  def keys(self):
    return list(self.iterkeys())

  def values(self):
    return self._dict.values()

  def iteritems(self):
    for key, value in self._dict.iteritems():
      yield key.key, value

  def iterkeys(self):
    for key in self._dict.keys():
      yield key.key

  def itervalues(self):
    return self._dict.itervalues()

  def get(self, key, *args):
    key = self.KeyWrapper(key, self.key_hash)
    return self._dict.get(key, *args)

  def setdefault(self, key, value):
    key = self.KeyWrapper(key, self.key_hash)
    return self._dict.setdefault(key, value)


class ValueIterableDict(abc.MutableMapping):
  """
  Just like any other map, but iterating over the map will yield the values
  instead of the keys. This is useful for lookup-maps where the values contain
  the keys that they are associated with in the map.
  """

  def __init__(self, iterable=None, map=None):
    if map is None:
      map = {}
    self._map = map
    if iterable is not None:
      self._map.update(iterable)

  def __repr__(self):
    return '{}({!r})'.format(type(self).__name__, self._map)

  def __len__(self):
    return len(self._map)

  def __iter__(self):
    return iter(self._map.values())

  def __getitem__(self, key):
    return self._map[key]

  def __setitem__(self, key, value):
    self._map[key] = value

  def __delitem__(self, key, value):
    del self._map[key]

  def __getattr__(self, attr):
    return getattr(self._map, attr)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
import types


class _InheritableModuleType(type):
  """
  Metaclass for a class type that wraps a module object and allow accessing
  members of that module or subclassing it.
  """

  def __new__(cls, name, bases, attrs):
    if bases == (types.ModuleType,):
      return type.__new__(cls, name, bases, attrs)
    for base in bases:
      if isinstance(base, _InheritableModuleType):
        module = base.__wrap_module__
        inheritable_type = base.__inheritable_type__
        break
    else:
      raise RuntimeError
    bases = tuple(x for x in bases if not isinstance(x, _InheritableModuleType))
    return type(name, (inheritable_type,) + bases, attrs)

  def __getattr__(self, name):
    return getattr(self.__wrap_module__, name)

  def __repr__(self):
    return '<inheritable ' + repr(self.__wrap_module__)[1:]


def create_inheritable_module(module, inheritable_type):
  """
  Wraps the specified *module* object and returns a class that can be
  subclassed to result in a direct subclass of *inheritable_type*.
  """

  return _InheritableModuleType(module.__name__, (types.ModuleType,),
    {'__wrap_module__': module, '__inheritable_type__': inheritable_type})


def make_inheritable(module_name, inheritable_type):
  """
  Makes the module with the specified *module_name* inheritable to the
  specified *inheritable_type*.
  """

  sys.modules[module_name] = create_inheritable_module(
    sys.modules[module_name], inheritable_type)




__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '2.1.1'


class NotSetType(object):

  def __new__(self):
    global NotSet
    if NotSet is not None:
      return NotSet
    NotSet = object.__new__(self)
    return NotSet

  def __repr__(self):
    return 'NotSet'

  def __bool__(self):
    return False

  def __nonzero__(self):
    return False


NotSet = None
NotSet = NotSetType()



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import types

from six.moves import range


class GenericMeta(type):
  """
  Metaclass that can be used for classes that need one or more datatypes
  pre-declared to function properly. The datatypes must be declared using
  the `__generic_args__` member and passed to the classes' `__getitem__()`
  operator to bind the class to these arguments.

  A generic class constructor may check it's `__generic_bind__` member to
  see if its generic arguments are bound or not.
  """

  def __init__(cls, *args, **kwargs):
    if not hasattr(cls, '__generic_args__'):
      raise TypeError('{}.__generic_args__ is not set'.format(cls.__name__))
    had_optional = False
    for index, item in enumerate(cls.__generic_args__):
      if not isinstance(item, tuple):
        item = (item,)
      arg_name = item[0]
      arg_default = item[1] if len(item) > 1 else NotImplemented
      if arg_default is NotImplemented and had_optional:
        raise ValueError('invalid {}.__generic_args__, default argument '
          'followed by non-default argument "{}"'
          .format(cls.__name__, arg_name))
      cls.__generic_args__[index] = (arg_name, arg_default)
    super(GenericMeta, cls).__init__(*args, **kwargs)
    if not hasattr(cls, '__generic_bind__'):
      cls.__generic_bind__ = None
    elif cls.__generic_bind__ is not None:
      assert len(cls.__generic_args__) == len(cls.__generic_bind__)
      for i in range(len(cls.__generic_args__)):
        value = cls.__generic_bind__[i]
        if isinstance(value, types.FunctionType):
          value = staticmethod(value)
        setattr(cls, cls.__generic_args__[i][0], value)
    else:
      collected_args = []
      missing = []
      for i, (arg_name, arg_default) in enumerate(cls.__generic_args__):
        if hasattr(cls, arg_name):
          collected_args.append(getattr(cls, arg_name))
        elif arg_default is not NotImplemented:
          collected_args.append(arg_default)
        else:
          missing.append(arg_name)
      if collected_args and missing:
        raise RuntimeError('{}: no all Generic arguments satisfied by '
          'class members (missing {})'.format(cls.__name__, ','.join(missing)))
      if collected_args:
        cls.__generic_bind__ = collected_args

  def __getitem__(cls, args):
    cls = getattr(cls, '__generic_base__', cls)
    if not isinstance(args, tuple):
      args = (args,)
    if len(args) > len(cls.__generic_args__):
      raise TypeError('{} takes at most {} generic arguments ({} given)'
        .format(cls.__name__, len(cls.__generic_args__), len(args)))
    # Find the number of required arguments.
    for index in range(len(cls.__generic_args__)):
      if cls.__generic_args__[index][1] != NotImplemented:
        break
    else:
      index = len(cls.__generic_args__)
    min_args = index
    if len(args) < min_args:
      raise TypeError('{} takes at least {} generic arguments ({} given)'
        .format(cls.__name__, min_args, len(args)))
    # Bind the generic arguments.
    bind_data = []
    for index in range(len(cls.__generic_args__)):
      arg_name, arg_default = cls.__generic_args__[index]
      if index < len(args):
        arg_value = args[index]
      else:
        assert arg_default is not NotImplemented
        arg_value = arg_default
      bind_data.append(arg_value)
    type_name = '{}[{}]'.format(cls.__name__, ', '.join(repr(x) for x in bind_data))
    data = {
      '__module__': cls.__module__,
      '__generic_bind__': bind_data,
      '__generic_base__': cls
    }
    return type(type_name, (cls,), data)


class _GenericHelperMeta(type):

  def __getitem__(self, args):
    if not isinstance(args, tuple):
      args = (args,)
    data = {'__generic_args__': list(args)}
    return GenericMeta('Generic[{0}]'.format(args), (object,), data)


Generic = _GenericHelperMeta('Generic', (object,), {})


def is_initialized(generic):
  if not isinstance(generic, type):
    generic = type(generic)
  return generic.__generic_bind__ is not None


def assert_initialized(generic):
  if not isinstance(generic, type):
    generic = type(generic)
  if not is_initialized(generic):
    raise RuntimeError('Missing generic arguments for {}'.format(generic.__name__))




import io
import re
import setuptools

with io.open('src/nr/types/__init__.py', encoding='utf8') as fp:
  version = re.search(r"__version__\s*=\s*'(.*)'", fp.read()).group(1)

with io.open('README.md', encoding='utf8') as fp:
  readme = fp.read()

setuptools.setup(
  name = 'nr.types',
  version = version,
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'Provides a number os useful data types for day-to-day Python programming.',
  long_description = readme,
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/NiklasRosenstein/python-nr.types',
  license = 'MIT',
  packages = setuptools.find_packages('src'),
  package_dir = {'': 'src'},
  install_requires = ['six'],
  extras_require = {
    'test': ['pytest', 'pytest-cov', 'BeautifulSoup4']
  }
)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

__all__ = ['OrderedSet']

import collections

from . import abc


class OrderedSet(abc.MutableSet):

  def __init__(self, iterable=None):
    self._index_map = {}
    self._content = collections.deque()
    if iterable is not None:
      self |= iterable

  def __repr__(self):
    if not self._content:
      return '%s()' % (type(self).__name__,)
    return '%s(%r)' % (type(self).__name__, list(self))

  def __eq__(self, other):
    if isinstance(other, OrderedSet):
      return len(self) == len(other) and list(self) == list(other)
    return set(self) == set(other)

  def __contains__(self, key):
    return key in self._index_map

  def __len__(self):
    return len(self._content)

  def __getitem__(self, index):
    return self._content[index]

  def __iter__(self):
    return iter(self._content)

  def __reversed__(self):
    return reversed(self._content)

  def add(self, key):
    if key not in self._index_map:
      self._index_map[key] = len(self._content)
      self._content.append(key)

  def copy(self):
    return type(self)(self)

  def discard(self, key):
    if key in self._index_map:
      index = self._index_map.pop(key)
      del self._content[index]

  def pop(self, last=True):
    if not self._content:
      raise KeyError('set is empty')
    key = self._content.pop() if last else self._content.popleft()
    self._index_map.pop(key)
    return key

  def update(self, iterable):
    for x in iterable:
      self.add(x)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import absolute_import

import functools
import itertools
import six

from . import NotSet
from six.moves import range, filter as _filter, filterfalse as _filterfalse, zip_longest

_next = next


class _dualmethod(object):
  """
  A combination of #classmethod() and instance methods. Methods decorated
  with this class can be called from the class and instance level. The first
  parameter will then refer to the class or the instance, respectively.
  """

  def __init__(self, func):
    self.func = func
    functools.update_wrapper(self, func)

  def __get__(self, instance, owner):
    assert owner is not None
    if instance is not None:
      return functools.partial(self.func, owner, instance)
    return functools.partial(self.func, owner)


class stream(object):
  """
  A wrapper for iterables that provides the stream processor functions of
  this module in an object-oriented interface.
  """

  def __init__(self, iterable):
    self.iterable = iter(iterable)

  def __iter__(self):
    return iter(self.iterable)

  def __next__(self):
    return _next(self.iterable)

  def __getitem__(self, val):
    if isinstance(val, slice):
      return self.slice(val.start, val.stop, val.step)
    else:
      raise TypeError('{} object is only subscriptable with slices'.format(type(self).__name__))

  @_dualmethod
  def call(cls, iterable, *a, **kw):
    """
    Calls every item in *iterable* with the specified arguments.
    """

    return cls(x(*a, **kw) for x in iterable)

  @_dualmethod
  def map(cls, iterable, func, *a, **kw):
    """
    Iterable-first replacement of Python's built-in `map()` function.
    """

    return cls(func(x, *a, **kw) for x in iterable)

  @_dualmethod
  def flatmap(cls, iterable, func):
    """
    Same as #map() but flattens the result.
    """

    def generator():
      for x in iterable:
        for y in func(x):
          yield y
    return cls(generator())

  @_dualmethod
  def filter(cls, iterable, cond, *a, **kw):
    """
    Iterable-first replacement of Python's built-in `filter()` function.
    """

    return cls(x for x in iterable if cond(x, *a, **kw))

  @_dualmethod
  def unique(cls, iterable, key=None):
    """
    Yields unique items from *iterable* whilst preserving the original order.
    """

    if key is None:
      key = lambda x: x
    def generator():
      seen = set()
      seen_add = seen.add
      for item in iterable:
        key_val = key(item)
        if key_val not in seen:
          seen_add(key_val)
          yield item
    return cls(generator())

  @_dualmethod
  def chunks(cls, iterable, n, fill=None):
    """
    Collects elements in fixed-length chunks.
    """

    return cls(zip_longest(*[iter(iterable)] * n, fillvalue=fill))

  @_dualmethod
  def concat(cls, iterables):
    """
    Similar to #itertools.chain.from_iterable().
    """

    def generator():
      for it in iterables:
        for element in it:
          yield element
    return cls(generator())

  @_dualmethod
  def chain(cls, *iterables):
    """
    Similar to #itertools.chain.from_iterable().
    """

    def generator():
      for it in iterables:
        for element in it:
          yield element
    return cls(generator())

  @_dualmethod
  def attr(cls, iterable, attr_name):
    """
    Applies #getattr() on all elements of *iterable*.
    """

    return cls(getattr(x, attr_name) for x in iterable)

  @_dualmethod
  def item(cls, iterable, key, default=NotSet):
    """
    Applies `__getitem__` on all elements of *iterable*.
    """

    if default is NotSet:
      return cls(x[key] for x in iterable)
    else:
      return cls(x.get(key, default) for x in iterable)

  @_dualmethod
  def of_type(cls, iterable, types):
    """
    Filters using #isinstance().
    """

    return cls(x for x in iterable if isinstance(x, types))

  @_dualmethod
  def partition(cls, iterable, pred):
    """
    Use a predicate to partition items into false and true entries.
    Returns a tuple of two streams with the first containing all elements
    for which *pred* returned #False and the other containing all elements
    where *pred* returned #True.
    """

    t1, t2 = itertools.tee(iterable)
    return cls(_filterfalse(pred, t1)), cls(_filter(pred, t2))

  @_dualmethod
  def dropwhile(cls, iterable, pred):
    return cls(itertools.dropwhile(pred, iterable))

  @_dualmethod
  def takewhile(cls, iterable, pred):
    return cls(itertools.takewhile(pred, iterable))

  @_dualmethod
  def groupby(cls, iterable, key=None):
    return cls(itertools.groupby(iterable, key))

  @_dualmethod
  def slice(cls, iterable, *args, **kwargs):
    return cls(itertools.islice(iterable, *args, **kwargs))

  if six.PY2:
    @_dualmethod
    def next(cls, iterable):
      if isinstance(iterable, stream):
        return iterable.__next__()
      else:
        return _next(iter(iterable))
  else:
    @_dualmethod
    def next(cls, iterable):
      return _next(iter(iterable))

  @_dualmethod
  def length(cls, iterable):
    """
    Returns the number of items in an iterable.
    """

    iterable = iter(iterable)
    count = 0
    while True:
      try:
        _next(iterable)
      except StopIteration:
        break
      count += 1
    return count

  @_dualmethod
  def consume(cls, iterable, n=None):
    if n is not None:
      for _ in range(n):
        try:
          _next(iterable)
        except StopIteration:
          break
    else:
      while True:
        try:
          _next(iterable)
        except StopIteration:
          break
    return iterable

  @_dualmethod
  def collect(cls, iterable, collect_cls=None, *args, **kwargs):
    return (collect_cls or list)(iterable, *args, **kwargs)


import sys
_module = sys.modules[__name__]
sys.modules[__name__] = stream



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

__all__ = ['Field', 'CleanRecord', 'Record', 'create_record']

from . import NotSet, abc
from .maps import OrderedDict
from .meta import InlineMetaclassBase
from six import iteritems, itervalues

import inspect
import six
import sys
import types


class Field(object):
  """
  Represents a field of a #Record class. Fields have a name, type and
  a default value. If a field has no name, it is "unbound" and the name
  may be derived from the name the field was assigned to in the #Record
  subclass declaration.
  """

  _global_create_index = 0

  def __init__(self, type, default=NotSet):
    self.name = None
    self.type = type
    self.default = default
    self.create_index = Field._global_create_index
    Field._global_create_index += 1

  def __repr__(self):
    return 'Field(name={!r}, type={!r}, default={!r})'.format(
      self.name, self.type, self.default)

  def get_default(self):
    if self.default is NotSet:
      raise RuntimeError('Field.default is NotSet')
    if isinstance(self.default, types.LambdaType):
      return self.default()
    return self.default

  @classmethod
  def with_name(cls, name, *args, **kwargs):
    obj = cls(*args, **kwargs)
    obj.name = name
    return obj


def compile_fields(decl, field_type):
  """
  Compiles a fields declaration to an #OrderedDict mapping to #Field objects.
  A fields declaration can be a string, list or dictionary.

  For a string: Comma or whitespace separated.

  For a list:

  * Field
  * Tuple[str, Type] (name, type)
  * Tuple[str, Type, Union[LambdaType, Any]] (name, type, default)

  For a dictionary:

  * Field
  * Tuple[Type, Union[LambdaType, Any]] (type, default)
  * Type
  """

  is_ordered = True

  # Split string into list of field names.
  if isinstance(decl, str):
    decl = [x.strip() for x in decl.split(',' if ',' in decl else ' ')]

  # Convert a mapping to the list form.
  elif isinstance(decl, abc.Mapping):
    is_ordered = False
    new_decl = []
    for key, value in iteritems(decl):
      if isinstance(value, tuple):
        value = (key,) + value
      elif isinstance(value, Field):
        value.name = key
      else:
        value = (key, value)
      new_decl.append(value)
    decl = new_decl

  compiled_fields = OrderedDict()
  for item in decl:
    if isinstance(item, str):
      field = field_type.with_name(item, None, NotSet)
    elif isinstance(item, field_type):
      field = item
    elif isinstance(item, tuple):
      if len(item) == 1:
        name, type_, default = item[0], None, NotSet
      elif len(item) == 2:
        name, type_, default = item[0], item[1], NotSet
      elif len(item) == 3:
        name, type_, default = item
      else:
        raise ValueError('invalid tuple Field declaration: {!r}'.format(item))
      field = field_type.with_name(name, type_, default)
    elif isinstance(item, dict):
      field = field_type.with_name(**item)
    else:
      raise TypeError('unexpected Field declaration: {!r}'.format(item))
    if not field.name:
      raise ValueError('unbound Field found: {!r}'.format(field.name))
    if field.name in compiled_fields:
      raise ValueError('duplicate Field name: {!r}'.format(field.name))
    compiled_fields[field.name] = field

  if not is_ordered:
    ordered = sorted(compiled_fields.items(), key=lambda x: x[1].create_index)
    compiled_fields = OrderedDict(ordered)

  return compiled_fields


class CleanRecord(InlineMetaclassBase):
  """
  A base-class similar to #typing.NamedTuple, but mutable. Fields can be
  specified using Python 3.6 class-member annotations, by setting the
  `__annotations__` or `__fields__` member to a list of annotations or
  by declaring class-members as #Field objects.

  In the following example all four `Person` declarations are identical.

  ```python
  class Person(Record):
    mail: str
    name: str = lambda: random_name()
    age: int = 0

  class Persom(Record):
    __fields__ = [
      ('mail', str),
      ('name', str, lambda: random_name()),
      ('age', int, 0)
    ]

  class Person(Record):
    mail = Field(str)
    name = Field(str, lambda: random_name())
    age = Field(str, 0)

  Person = create_record('Person', [
    ('mail', str),
    Field.with_name('name', str, lambda: random_name()),
    ('age', int, 0)
  ])
  ```
  """

  __field_type__ = Field

  def __metainit__(self, name, bases, dict):
    """
    Overrides #InlineMetaclassBase.__metainit__(). Converts the fields or
    annotations defined on the class to a common dictionary representation.
    The source of the fields is determined in the following order:

    * Attribute `__fields__`
    * Attribute `__annotations__`
    * Member attributes (only #Field instances)
    """

    field_type = self.__field_type__

    # Determine the source for the field declaration.
    if '__fields__' in dict:
      fields = dict['__fields__']
    elif '__annotations__' in dict:
      fields = dict['__annotations__']
      if isinstance(fields, abc.Mapping):
        for key, value in iteritems(fields):
          fields[key] = (value, getattr(self, key, NotSet))
    else:
      fields = []
      for key, value in iteritems(dict):
        if isinstance(value, field_type):
          if value.name is None:
            value.name = key
          elif value.name != key:
            raise ValueError('Field.name mismatches key: {!r} != {!r}'
              .format(value.name, key))
          fields.append(value)
      fields.sort(key=lambda x: x.create_index)

    # Merge with parent class fields.
    mro_fields = OrderedDict()
    for base in reversed(bases):
      if hasattr(base, '__fields__'):
        mro_fields.update(base.__fields__)
    mro_fields.update(compile_fields(fields, field_type))

    for name, field in iteritems(mro_fields):
      if name != field.name:
        raise ValueError('mismatching Field name: name({!r}) != Field.name({!r})'
          .format(name, field.name))

    self.__fields__ = mro_fields

  def __init__(self, *args, **kwargs):
    type_name = type(self).__name__

    # Validate number of arguments.
    nargs = len(args) + len(kwargs)
    if nargs > len(self.__fields__):
      raise TypeError('{}() expected at most {} arguments, got {}'
        .format(type_name, len(self.__fields__), nargs))

    # Raise an exception for any unknown keyword arguments.
    for key in kwargs:
      if key not in self.__fields__:
        raise TypeError('{}() unexpected keyword argument "{}"'
          .format(type_name, key))

    # Map positional arguments to keyword arguments.
    for arg, (key, field) in zip(args, iteritems(self.__fields__)):
      if field.name in kwargs:
        raise TypeError('{}() got duplicate argument "{}"'
          .format(type_name, name))
      kwargs[field.name] = arg

    # Create attributes.
    for key, field in iteritems(self.__fields__):
      value = kwargs.get(key, NotSet)
      if value is NotSet:
        if field.default is NotSet:
          raise TypeError('{}() missing argument "{}"'.format(type_name, key))
        value = field.get_default()
      setattr(self, key, value)

  def __repr__(self):
    values = ((k, getattr(self, k)) for k in self.__fields__)
    members = ', '.join('{}={!r}'.format(k, v) for k, v in values)
    return '{}({})'.format(type(self).__name__, members)

  def __eq__(self, other):
    if isinstance(other, CleanRecord) and self.__fields__ is other.__fields__:
      for key in self.__fields__:
        if getattr(self, key) != getattr(other, key):
          return False
      return True
    return False


class ToJSON(object):
  """
  A mixin for the #CleanRecord class that adds a #to_json() method.
  Different from the #as_dict() method in the #AsDict mixin, #to_json()
  is called recursively on any of the attributes if they have a #to_json()
  method.
  """

  def to_json(self):
    """
    Converts the record to a representation that can be dumped into JSON
    format. For any member, it will check if that member has a `to_json()`
    method and call it. Mappings and sequences are converted recursively.

    Note that this method does not guarantee that the returned object will
    be JSON serializable afterwards. For special cases, the method should
    be overwritten.
    """

    def coerce(value):
      if hasattr(value, 'to_json'):
        return value.to_json()
      elif isinstance(value, abc.Mapping):
        return dict((k, value[k]) for k in value)
      elif isinstance(value, abc.Sequence) and not isinstance(value, (six.string_types, six.binary_type, bytearray)):
        return [coerce(x) for x in value]
      else:
        return value

    result = {}
    for key in self.__fields__:
      result[key] = coerce(getattr(self, key))
    return result


class AsDict(object):
  """
  A mixin for the #CleanRecord class that adds an #as_dict() method.
  """

  def as_dict(self):
    return dict((k, getattr(self, k)) for k in self.__fields__)


class Sequence(object):
  """
  A mixin for the #CleanRecord class that implements the mutable sequence
  interface.
  """

  def __iter__(self):
    for key in self.__fields__:
      yield getattr(self, key)

  def __len__(self):
    return len(self.__fields__)

  def __getitem__(self, index):
    if hasattr(index, '__index__'):
      return getattr(self, self.__fields__[index.__index__()].name)
    elif isinstance(index, str):
      return getattr(self, str)
    else:
      raise TypeError('cannot index with {} object'
        .format(type(index).__name__))

  def __setitem__(self, index, value):
    if hasattr(index, '__index__'):
      setattr(self, self.__fields__[index.__index__()].name, value)
    elif isinstance(index, str):
      setattr(self, index, value)
    else:
      raise TypeError('cannot index with {} object'
        .format(type(index).__name__))


class Record(CleanRecord, ToJSON, AsDict):
  pass


def _get_calling_module_name(_stackdepth=0):
  frame = sys._getframe(_stackdepth+2)
  return frame.f_globals.get('__name__', __name__)


def create_record(name, fields, *mixins, **kwargs):
  """
  Creates a new #Record subclass. If at least one *mixin* is specified, it is
  mixed into the parent class of the created #Record subclass. One of the
  mixins can also be a separate #Record subclass in which case it is used
  as the parent class instead of #Record.
  """

  module = kwargs.pop('module', None)
  if module is None:
    module = _get_calling_module_name()

  for key in kwargs:
    raise TypeError('unexpected keyword argument: {!r}'.format(key))

  base = next((x for x in mixins if issubclass(x, CleanRecord)), None)
  if not base:
    mixins = mixins + (Record,)

  return type(name, mixins, {'__fields__': fields, '__module__': module})


def clone_record(record_cls, name=None, module=None):
  """
  Creates a clone of the specified *record_cls*. Optionally overrides the
  new classes *name*.
  """

  if module is None:
    module = _get_calling_module_name()

  assert isinstance(record_cls, type), record_cls
  assert issubclass(record_cls, CleanRecord), record_cls

  name = name or record_cls.__name__
  return create_record(name, record_cls.__fields__, *record_cls.__bases__,
    module=module)


create = create_record


from . import moduletools as _moduletools
_moduletools.make_inheritable(__name__, Record)
